from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import argparse
from tqdm import tqdm


argparser = argparse.ArgumentParser(description="Split T2T reference into individual chromosomes")
argparser.add_argument("-g",'--genome', help="Path to the reference fasta file", type=str, default="~/ref/T2T.fa")
argparser.add_argument('-n','--name',help="Name of the reference", type = str, default="CHM13.T2T")
argparser.add_argument('-d','--directory',help="Path to the output directory", type = str, default="./data/ref")
args = argparser.parse_args()

def extract_headers_and_sequences(fasta_file:str,ref_name:str = "CHM13.T2T", dir=None) -> tuple[list[str], list[str], list[str]]:
    headers = []
    sequences = []

    if dir:
        ref_name = f"{dir}/{ref_name}"

    for record in SeqIO.parse(fasta_file, "fasta"):
        headers.append(record.description)
        sequences.append(record.seq)

    chr_ids = [x.split(" ")[6].strip(',') for x in headers]
    file_names = [f"{ref_name}.chromosome.{x}.fa" for x in chr_ids]

    return file_names, headers, sequences


def save_fasta_files(file_names:list[str], headers:list[str], sequences:list[str]) -> None:


    for i in tqdm(range(len(file_names))):
        sequence = Seq(sequences[i])
        record = SeqRecord(sequence, id=headers[i])
        with open(file_names[i], "w") as f:
            SeqIO.write(record, f, "fasta")
    return


if __name__ == "__main__":
    ref = 'GCF_009914755.1_T2T-CHM13v2.0_genomic.fna'
    save_fasta_files(
        *extract_headers_and_sequences(
            fasta_file=args.genome,
            dir=args.directory,
            ref_name=args.name,
            )
        )



#!/usr/bin/python3

from pickle import FALSE, TRUE
import numpy as np
import multiprocessing as mp
from joblib import Parallel, delayed
from tabulate import tabulate as tab
import os
from tqdm import tqdm
import pandas as pd 
import sys
import contextlib
import joblib
import time
from tqdm import tqdm

types = {'DNA' : 'Reference', 'RNA' : 'RevCompl', '-RNA' : 'Reference'}

@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()



def consecblock(oligo,L,l,hdist_grouped,
                maxccmatch:int=0, # largest possible match
                currentccmatch:int=0, # current match block
                prev:int=1, # pointer to previous sublength
                )->int:
    # find consecutive 0 in each L-mer

    for sub in range(L-l+1):         # check the homology of all consecutive l-mers
        current = hdist_grouped[oligo,sub]
        if(current == 0 and prev>0):     # first zero in a new block
            currentccmatch = l 
        elif(current == 0 and prev==0):     # match block continues
            currentccmatch = currentccmatch + 1
        elif(current > 0 and prev==0):       # end of previous block    
            maxccmatch = max([maxccmatch,currentccmatch])
            currentccmatch = 0
        else:
            currentccmatch = 0

        #print(f'Oligo: '+str(oligo)+', current HUSH: '+str(current)+', current block: '+str(currentccmatch)+', max: '+str(maxccmatch)+', prev: '+str(prev))
        prev = current
    
    maxccmatch = max([maxccmatch,currentccmatch])
    
    return maxccmatch


def reform_hush_combined(nt_type:str='DNA',
                         L:int=40,
                         l:int=22,
                         currentfolder:os.PathLike = './data',
                         until:int=3)->None:
    
    suffix = types[nt_type]
    roilist = currentfolder+'/rois/all_regions.tsv'
    rd = pd.read_csv(roilist,sep="\t",header=0)
    ROIcount = len(rd)

    infolder = currentfolder+'/candidates/'
    out = currentfolder+'/HUSH_candidates/'

    try:
        os.mkdir(out)
    except FileExistsError:
        print("Saving to existing 'HUSH_candidates' directory.")

    ## each oligo gets an off-target score based on:
    # 1. The longest consecutive perfect off-target match (consecutive sublength oligos returning 0 in nHUSH results)
    # 2. The sum of mismatch counts in sublength oligos. Challenging to interpret but higher for central mismatches than closer to the edges.
    # Encode both scores to retrieve them separately after building db.

    for roi in tqdm(range(ROIcount),'Processing all regions'):
        
        filename = 'roi_'+str(rd.window_id[roi])+'.GC35to85_'+suffix+'.fa'
        fasta = infolder+filename
        hush = fasta+'.'+str(l)+'mers.nh.L'+str(l)+'.mindist.uint8'
        hdist = np.fromfile(hush,'uint8')

        # correct for aberrant values after nHUSH.
        hdist[hdist>until] = until+1
        n = len(hdist)

        hdist_grouped = hdist.reshape([n//(L-l+1), (L-l+1)])
        hdist_grouped_sum = hdist.reshape([n//(L-l+1), (L-l+1)]).sum(axis=1)            

        with tqdm_joblib(tqdm(desc="Processing all oligos in region "+str(rd.window_id[roi]), total=len(hdist_grouped))) as progress_bar:
            results = Parallel(n_jobs=40)(delayed(consecblock)(oligo,L,l,hdist_grouped) for oligo in range(len(hdist_grouped)))

        # write results into fasta file  
        f = open(fasta,'r')
        full = f.read()
        splitseq = full.splitlines()

        for k in range(len(hdist_grouped)):
            splitseq[2*k] = splitseq[2*k] + "\n"
            # /!\ UGLY SOLUTION
            # encode the two mismatch parameters (min mismatch hit and sum mismatch count) in a format compatible with ifpd2 db make
            splitseq[2*k+1] = splitseq[2*k+1] + ", " + "111" + str(results[k]) + "987" + str(hdist_grouped_sum[k]) + "\n"
                
        o = open(out+filename,'w')
        outseq = ''.join(splitseq)
        o.write(outseq)
        o.close()
        f.close()

        return

if __name__ == "__main__":
    #syntax: ./reform_hush_combined.py DNA/RNA/-RNA 40 22 3 

    print(f'Number of arguments: '+str(len(sys.argv)))
    if(len(sys.argv) < 5):
        print(f'The combined scoring approach requires nHUSH to be run using sublength oligos.\n')
        print(f'Required arguments: [DNA/RNA/-RNA] [length] [sublength] [until].\n')
        print(f'With "until" the same parameter as used when running nHUSH.\n')
        print(f'Incorrect number of arguments. Exiting...')
        exit(-1)

    # types = {'DNA' : 'Reference', 'RNA' : 'RevCompl', '-RNA' : 'Reference'}
    # suffix = types[sys.argv[1]]

    L = int(sys.argv[2])     #full oligo length
    print(f'Length: '+str(L))
    l = int(sys.argv[3])    # sublength used for nHUSH
    print(f'Sublength: '+str(l))
    until = int(sys.argv[4])
    print(f'HUSH was run until '+str(until)+' mismatches.')
    # call the function here
    reform_hush_combined(nt_type=sys.argv[1],L=L,l=l,until=until)  



#!/usr/bin/python3

# GG's code sniplet (R)
#max_dg_fraction = 0.5
#max_off_targets = 99

#db[, off_target_score := 1-off_target_no/max_off_targets]
#db[off_target_score <= 0, off_target_score := 0]
#
#db[ss_dG/Tm_dG > max_dg_fraction, ss_score := 0]
#db[ss_dG/Tm_dG <= max_dg_fraction, ss_score := 1-ss_dG/Tm_dG/max_dg_fraction]
#db[ss_dG >= 0, ss_score := 1]
#
#db[, score := (off_target_score+ss_score)/2]

import sys
import re
from tqdm import tqdm 

def score_gg(d):
    # Extract variables from the dictionary
    ss_dG = float(d['ss_dG'])
    off_target_no = int(d['off_target_no'])
    Tm_dG = float(d['Tm_dG'])
    # Parameters
    max_dg_fraction = 0.5
    max_off_targets = 99
    # Calculate the score
    off_target_score = 1-off_target_no/max_off_targets
    if off_target_score < 0:
        off_target_score = 0
    ss_score = 0
    if(ss_dG/Tm_dG <= max_dg_fraction):
        ss_score = 1-ss_dG/Tm_dG/max_dg_fraction

    if ss_dG >= 0:
        ss_score = 1
    score = 1-(off_target_score + ss_score)/2
    return score

def score_gg_nhush(d):
    # Extract variables from the dictionary
    ss_dG = float(d['ss_dG'])
    off_target_no = int(d['off_target_no'])
    Tm_dG = float(d['Tm_dG'])
    # Parameters
    max_dg_fraction = 0.5
    max_off_targets = 99
    # Calculate the score
    # 0 -> 1, 1->1/2, 2->1/3 etc
    off_target_score = 1/(1+off_target_no)
    if off_target_score < 0:
        off_target_score = 0
    ss_score = 0
    if(ss_dG/Tm_dG <= max_dg_fraction):
        ss_score = 1-ss_dG/Tm_dG/max_dg_fraction

    if ss_dG >= 0:
        ss_score = 1
    #score = (off_target_score + ss_score)/2
    score = 1-0.5*ss_score + 0.5*off_target_score

    return score

def score_q_nhush(d):
    # Extract variables from the dictionary
    ss_dG = float(d['ss_dG'])
    off_target_no = int(d['off_target_no'])
    off_target_sum = int(d['off_target_sum'])
    Tm_dG = float(d['Tm_dG'])
    Tm = float(d['Tm'])
    seq = str(d['sequence'])
    # Parameters
    max_dg_fraction = 0.3
    max_offtarget = 4   # correction for aberrant values after mindist
    # Calculate the score
    # 0 -> 1, 1->1/2, 2->1/3 etc
    if(off_target_no>max_offtarget):
        off_target_score = 1e99
    else:    
        off_target_score = 1/(1+off_target_no)**2
    if(off_target_sum==0):
        sum_target_score = 1e99
    else:
        sum_target_score = 10/(1+off_target_sum)**(1/2)      
    ss_score = 1
    if(ss_dG/Tm_dG <= max_dg_fraction):
        ss_score = ss_dG/Tm_dG/max_dg_fraction
    if ss_dG >= 0:
        ss_score = 0
    if re.search(r'((\w)\2{5,})',seq):
        ss_score = 1e99
    tm_score = ((68-Tm)/68)**2    
    #score = (off_target_score + ss_score)/2
    score = 0.2*ss_score + 0.2*tm_score + 0.6*off_target_score + sum_target_score

    return score    

def score_q_consec(d,max_off_targets,maxid):
    # Extract variables from the dictionary
    ss_dG = float(d['ss_dG'])
    off_target_no = int(d['off_target_no'])
    Tm_dG = float(d['Tm_dG'])
    seq = str(d['sequence'])
    # Parameters
    max_dg_fraction = 0.5
    # Calculate the score
    if off_target_no > max_off_targets:
        off_target_score = 1e99
    else:
        off_target_score = off_target_no/max_off_targets
    ss_score = 1
    if(ss_dG/Tm_dG <= max_dg_fraction):
        ss_score = ss_dG/Tm_dG/max_dg_fraction
    if ss_dG >= 0:
        ss_score = 0
    query = r'((\w)\2{'+str(maxid-1)+',})'
    if re.search(query,seq):
        ss_score = 1e99

    score = (off_target_score + ss_score)/2
    return score

def score_q_combined(d,max_consec,maxid,targetTemp):         
   
    # Extract variables from the dictionary
    ss_dG = float(d['ss_dG'])
    off_target_no = int(d['off_target_no'])    # actually the longest consecutive stretch
    off_target_sum = int(d['off_target_sum'])  # cumulated mismatch counts
    Tm_dG = float(d['Tm_dG'])
    Tm = float(d['Tm'])
    seq = str(d['sequence'])

    # Parameters
    max_dg_fraction = 0.5

    # Calculate the oligo cost
    # 1. Secondary structures
    ss_cost = 1e10
    if ss_dG >= 0:
        ss_cost = 0
    elif(ss_dG/Tm_dG <= max_dg_fraction):
        ss_cost = ss_dG/Tm_dG/max_dg_fraction       # between 0-1 

    # 2. Melting temperature
    # Build into escafish in the future to minimize the Tm spread at the probe level?
    tm_cost = ((targetTemp-Tm)/5)**2  # +-5 deg should give at 0-1 range

    # 3. Off-target homology. Consecutive stretches
    if off_target_no > max_consec:
        consec_cost = 1e10
    else:
        consec_cost = off_target_no/max_consec     # between 0-1

    # 4. Off-target homology. Cumulated Hamming distances
    # The hard filter on consecutive stretches restrict the range of values to be expected here.
    sum_offtarget_cost = (10/(1+off_target_sum))**(1/2)     # between 0-1 (min value for a central mismatch in a 30mer: 9)

    # 5. Homopolymers
    query = r'((\w)\2{'+str(maxid-1)+',})'      
    if re.search(query,seq):
        ss_cost = 1e10

    # Regroup into single oligo cost
    score = 0.2*ss_cost + 0.2*tm_cost + 0.6*consec_cost + 0.2*sum_offtarget_cost

    return score    

def score_q_combined_bl(d,max_consec,maxid,targetTemp,hamdist):         
   
    # Extract variables from the dictionary
    ss_dG = float(d['ss_dG'])
    off_target_no = int(d['off_target_no'])    # actually the longest consecutive stretch
    off_target_sum = int(d['off_target_sum'])  # cumulated mismatch counts
    Tm_dG = float(d['Tm_dG'])
    Tm = float(d['Tm'])
    seq = str(d['sequence'])
    distBL = int(d['bl_dist'])

    # Parameters
    max_dg_fraction = 0.5

    # Calculate the oligo cost
    # 1. Secondary structures
    ss_cost = 1e10
    if ss_dG >= 0:
        ss_cost = 0
    elif(ss_dG/Tm_dG <= max_dg_fraction):
        ss_cost = ss_dG/Tm_dG/max_dg_fraction       # between 0-1 

    # 2. Melting temperature
    # Build into escafish in the future to minimize the Tm spread at the probe level?
    tm_cost = ((targetTemp-Tm)/5)**2  # +-5 deg should give at 0-1 range

    # 3. Off-target homology. Consecutive stretches
    if off_target_no > max_consec:
        consec_cost = 1e10
    else:
        consec_cost = off_target_no/max_consec     # between 0-1

    # 4. Off-target homology. Cumulated Hamming distances
    # The hard filter on consecutive stretches restrict the range of values to be expected here.
    if distBL <= hamdist:
        sum_offtarget_cost = 1e10
    else:
        sum_offtarget_cost = (10/(1+off_target_sum))**(1/2)     # between 0-1 (min value for a central mismatch in a 30mer: 9)

    # 5. Homopolymers
    query = r'((\w)\2{'+str(maxid-1)+',})'      
    if re.search(query,seq):
        ss_cost = 1e10

    # Regroup into single oligo cost
    score = 0.2*ss_cost + 0.2*tm_cost + 0.6*consec_cost + 0.2*sum_offtarget_cost

    return score    


score_functions = {'gg' : score_gg,
                   'gg_nhush': score_gg_nhush,
                   'q': score_q_nhush,
                   'q_cc': score_q_consec,
                   'q_combined': score_q_combined,
                   'q_bl': score_q_combined_bl}

if __name__ == '__main__':
    header = sys.stdin.readline().strip()

    if(len(sys.argv) < 2):
        print(f'No score function selected')
        print(f'Available:')
        for key in score_functions:
            print(f'{key}')
        exit(-1)

    score_function = score_functions[sys.argv[1]]

    # unless specified, allow max 6 consecutive identical base pairs
    maxid = "6"
    if(len(sys.argv) == 3):
        maxmm = int(sys.argv[2])
    elif(len(sys.argv) == 5):
        maxmm = int(sys.argv[2])
        maxid = int(sys.argv[3])
        targetTemp = int(sys.argv[4])
    elif(len(sys.argv) == 6):
        maxmm = int(sys.argv[2])
        maxid = int(sys.argv[3])
        targetTemp = int(sys.argv[4])    
        hamdist = int(sys.argv[5]) 
    print(f"{header}\toligo_cost")

    for line in tqdm(sys.stdin,"Reading database"):
        line = line.strip()
        d = dict(zip(header.split('\t'), line.split('\t')))

        if (sys.argv[1] == 'q_cc'):
            score = score_function(d,maxmm,maxid)
        elif (sys.argv[1] == 'q_combined'):
            score = score_function(d,maxmm,maxid,targetTemp)
        elif (sys.argv[1] == 'q_bl'):
            score = score_function(d,maxmm,maxid,targetTemp,hamdist)
        else: 
            score = score_function(d)
        # Optimization will not work if score < 0
        if(score < 0):
            score = 0

        print(f"{line}\t{score}")



#!/usr/bin/python3

from pickle import FALSE, TRUE
import numpy as np
import multiprocessing as mp
from joblib import Parallel, delayed
from tabulate import tabulate as tab
import os
from tqdm import tqdm
import pandas as pd
import sys
import contextlib
import joblib
import time
from tqdm import tqdm

types = {'DNA' : 'Reference', 'RNA' : 'RevCompl', '-RNA' : 'Reference'}

@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()



def consecblock(oligo,L,l,hdist_grouped,
                maxccmatch = 0, # largest possible match,
                currentccmatch = 0,  # current match block,
                prev = 1, # pointer to previous sublength
                )->int:
    # find consecutive 0 in each L-mer

    for sub in range(L-l+1):         # check the homology of all consecutive l-mers
        current = hdist_grouped[oligo,sub]
        if(current == 0 and prev>0):     # first zero in a new block
            currentccmatch = l 
        elif(current == 0 and prev==0):     # match block continues
            currentccmatch = currentccmatch + 1
        elif(current > 0 and prev==0):       # end of previous block    
            maxccmatch = max([maxccmatch,currentccmatch])
            currentccmatch = 0
        else:
            currentccmatch = 0

        #print(f'Oligo: '+str(oligo)+', current HUSH: '+str(current)+', current block: '+str(currentccmatch)+', max: '+str(maxccmatch)+', prev: '+str(prev))
        prev = current
    
    maxccmatch = max([maxccmatch,currentccmatch])
    
    return maxccmatch



def reform_hush_consec(nt_type:str='DNA',
                       sub:bool=False,
                       L:int=80,
                       l:int=22,
                       currentfolder:os.PathLike = './data',
                       )->None:


    suffix = types[nt_type]

    roilist = currentfolder+'/rois/all_regions.tsv'
    rd = pd.read_csv(roilist,sep="\t",header=0)
    ROIcount = len(rd)

    infolder = currentfolder+'/candidates/'
    out = currentfolder+'/HUSH_candidates/'

    try:
        os.mkdir(out)
    except FileExistsError:
        print("Saving to existing 'HUSH_candidates' directory.")

    if (sub==TRUE):
        
        for roi in tqdm(range(ROIcount),'Processing all regions'):
            ## each oligo gets an off-target score based on the max number of consecutive perfect matches (0 in nHUSH results)
            # and the resulting sum of such consecutive blocks
            # one 0 => one sublength perfect match
            # each consecutive 0 => +1 to length
            # count each block separately and sum the results
            
            filename = 'roi_'+str(rd.window_id[roi])+'.GC35to85_'+suffix+'.fa'
            fasta = infolder+filename
            hush = fasta+'.'+str(l)+'mers.nh.L'+str(l)+'.mindist.uint8'
            hdist = np.fromfile(hush,'uint8')

            n = len(hdist)

            hdist_grouped = hdist.reshape([n//(L-l+1), (L-l+1)])
            

            with tqdm_joblib(tqdm(desc="Processing all oligos in region "+str(rd.window_id[roi]), total=len(hdist_grouped))) as progress_bar:
                results = Parallel(n_jobs=40)(delayed(consecblock)(oligo,L,l,hdist_grouped) for oligo in range(len(hdist_grouped)))
            #for oligo in range(590):
            #    consecblock(oligo,L,l,hdist_grouped)

            #time.sleep(20000)

            # write results into fasta file  
            f = open(fasta,'r')
            full = f.read()
            splitseq = full.splitlines()

            for k in range(len(hdist_grouped)):
                #print('Current row: '+str(k2))
                splitseq[2*k] = splitseq[2*k] + "\n"
                # /!\ UGLY SOLUTION
                # encode the two mismatch parameters (min mismatch hit and sum mismatch count) in a format compatible with ifpd2 db make
                splitseq[2*k+1] = splitseq[2*k+1] + ", " + str(results[k]) + "\n"
                    
            o = open(out+filename,'w')
            outseq = ''.join(splitseq)
            o.write(outseq)
            o.close()
            f.close()  

    # else:
    #     # each oligo gets an off-target score based on the min number of mismatches as found by nHUSH

    #     for k1 in tqdm(range(ROIcount), 'Processing all ROI'):
    #         filename = 'roi_'+str(k1+1)+'.GC35to85_'+suffix+'.fa'
    #         fasta = infolder+filename
    #         hush = fasta+'.nh.L'+str(L)+'.mindist.uint8'
    #         hdist = np.fromfile(hush,'uint8')

    #         n = len(hdist)                    
    #         f = open(fasta,'r')
    #         full = f.read()
    #         splitseq = full.splitlines()

    #         for k2 in range(n):
    #             #print('Current row: '+str(k2))
    #             splitseq[2*k2] = splitseq[2*k2] + "\n"
    #             # /!\ UGLY SOLUTION
    #             # encode the two mismatch parameters (min mismatch hit and sum mismatch count) in a format compatible with ifpd2 db make
    #             splitseq[2*k2+1] = splitseq[2*k2+1] + ", " + "111" + str(hdist[k2]) + "98799"  + "\n"
                    
    #         o = open(out+filename,'w')
    #         outseq = ''.join(splitseq)
    #         o.write(outseq)
    #         o.close()
    #         f.close()         
    return

if __name__ == "__main__":
    # for Command Line use
    #syntax: ./reform_hush.py DNA/RNA/-RNA 80 (22)

    nt_type = sys.argv[1] # DNA/RNA/-RNA
    print(f'Number of arguments: '+str(len(sys.argv)))
    if(len(sys.argv) == 3):
        # no sublength was used
        sub = FALSE
    elif(len(sys.argv) == 4):
        sub = TRUE
    else:
         print(f'Incorrect number of arguments. Exiting...')
         exit(-1)

    L = int(sys.argv[2])     #full oligo length
    print(f'Length: '+str(L))
    if (sub==TRUE):
        l = int(sys.argv[3])
        print(f'Sublength: '+str(l))
        reform_hush_consec(nt_type=nt_type,sub=sub,L=L,l=l)
    else:
        print(f'No sublength used')
        reform_hush_consec(nt_type=nt_type,sub=sub,L=L)
    
    


    